from ywh2bt.utils import get_all_subclasses
from ywh2bt.exceptions import BugTrackerNotFound, MultipleBugTrackerMacth
from .abstract import ConfigObject, BugTrackerConfig
from .yeswehack import YesWeHackConfig
from .packages import ExtraPackageConfig
import os
from pathlib import Path
import sys
import yaml
from ywh2bt.utils import read_input
from ywh2bt.logging import logger
from colorama import Fore, Style

__all__ = ["GlobalConfig"]

"""
Module to load configuration file globaly
"""


class GlobalConfig(ConfigObject):

    """
    Configuration file representation

    :attr list mandatory_keys: keys needed in config_file
    :attr str name: ConfigObject identifier
    """

    mandatory_keys = ["yeswehack", "bugtrackers"]
    name = "global"

    ############################################################
    ####################### Constructor ########################
    ############################################################
    def __init__(
        self, configure_mode=False, no_interactive=False, filename=None
    ):
        self.default_supported_bugtrackers = []
        self.no_interactive = no_interactive
        self.configure_mode = configure_mode
        self.filename = (
            Path(filename) if filename else self.get_default_config_file()
        )
        config = self._load(self.filename)
        self.bugtrackers = []
        self.yeswehack = []
        self.packages = []
        if config or not configure_mode:
            self._build(config)
        if configure_mode:
            self.configure()

    ############################################################
    ################### Instance methods #######################
    ############################################################

    def configure(self):
        """
        Load or setup configuration, depends of attribute and existence configuration keys.
        """
        logger.info("Welcome in ywh2bt configuration tools")

        if self.no_interactive:
            logger.warning(
                "You have choose non interactive mode ! All information you will write will be stored clearly visible."
            )
        exit_config = False
        if self.yeswehack or self.bugtrackers:
            logger.info(
                "Existing configuration {} loaded".format(self.filename)
            )
        while not exit_config:
            configure_new_elem = read_input(
                Fore.BLUE + "Configure new element ? [y/N]: " + Style.RESET_ALL
            )
            if configure_new_elem in ["y", "Y"]:
                try:
                    self.configure_new_packages()
                    self.configure_new_bugtrackers()
                    self.configure_new_yeswehack()
                except Exception as e:
                    logger.error(str(e))
            elif configure_new_elem in ["n", "N", ""]:
                exit_config = True
        self._update_configuration()
        self.write()

    def configure_new_bugtrackers(self):
        """
        Add new bugtrackers with user interaction
        """
        counter = len(self.bugtrackers) + 1
        exit_config = False
        while not exit_config:
            exit_configuration = read_input(
                Fore.BLUE + "Add new BugTracker ? [y/N]: " + Style.RESET_ALL
            )
            if exit_configuration in ["n", "N", ""]:
                exit_config = True
            elif exit_configuration in ["y", "Y"]:
                logger.info(
                    "Supported engine are: "
                    + Fore.YELLOW
                    + ", ".join(
                        [
                            bt_type.title()
                            for bt_type in self.default_supported_bugtrackers
                        ]
                    )
                    + Style.RESET_ALL
                )
                while True:
                    tracker_type = read_input(
                        Fore.BLUE + "Type: " + Style.RESET_ALL
                    ).lower()
                    if not tracker_type in self.default_supported_bugtrackers:
                        logger.error("Unsuported type")
                    else:
                        break
                self.bugtrackers.append(
                    self.get_bugtracker_class(tracker_type)(
                        "bugtracker_{}".format(counter),
                        no_interactive=self.no_interactive,
                        configure_mode=True,
                    )
                )
                counter += 1

    def configure_new_packages(self):
        """
        Add new packages with user interaction
        """
        exit_config = False
        while not exit_config:
            exit_configuration = read_input(
                Fore.BLUE
                + "Add Package configuration element ? [y/N]: "
                + Style.RESET_ALL
            )
            if exit_configuration in ["n", "N", ""]:
                exit_config = True
            elif exit_configuration in ["y", "Y"]:
                self.packages.append(ExtraPackageConfig(configure_mode=True))
        self.default_supported_bugtrackers = [
            tr.bugtracker_type for tr in get_all_subclasses(BugTrackerConfig)
        ] or []

    def configure_new_yeswehack(self):
        """
        Add new YesWeHack with user interaction
        """
        exit_config = False
        count = len(self.yeswehack) + 1
        while not exit_config:
            exit_configuration = read_input(
                Fore.BLUE
                + "Add YesWeHack configuration element ? [y/N]: "
                + Style.RESET_ALL
            )
            if exit_configuration in ["n", "N", ""]:
                exit_config = True
            elif exit_configuration in ["y", "Y"]:
                self.yeswehack.append(
                    YesWeHackConfig(
                        "yeswehack_{}".format(count),
                        self.bugtrackers,
                        no_interactive=self.no_interactive,
                        configure_mode=True,
                    )
                )

    def get_default_config_file(self):
        """
        Return default configuration file.
        """
        if os.name == "posix":
            config_dir = Path(os.environ.get("HOME") + "/")
        else:
            config_dir = Path(os.environ.get("APPDATA") + "/ywh2bt/")
        if not config_dir.exists():
            config_dir.mkdir(parents=True)
        config_file = Path(str(config_dir) + "/.ywh2bt.cfg")
        return config_file

    def to_dict(self):
        """
        Map object to dictionary
        """
        bts = {}
        for bugtracker in self.bugtrackers:
            bts = {**bts, **bugtracker.to_dict()}
        ywhs = {}
        for ywh in self.yeswehack:
            ywhs = {**ywhs, **ywh.to_dict()}
        component = {"yeswehack": ywhs, "bugtrackers": bts}
        if self.packages:
            pkgs = [pkg.to_dict() for pkg in self.packages]
            component["packages"] = pkgs
        return component

    def write(self):
        """
        Write GlobalConfig to configuration file.
        """
        yaml_configuration = yaml.dump(self.to_dict())
        with open(self.filename, "w") as f:
            f.write(yaml_configuration)

    def _build(self, config):
        """
        Build object from config dictionary information.

        :param dict config: configuration.
        """
        super().__init__(self.mandatory_keys, self.name, **config)
        self.bugtrackers = self._config_bugtrackers(
            config["bugtrackers"], configure_mode=False
        )
        self.yeswehack = self._config_ywh(config["yeswehack"])
        if "packages" in config:
            self.packages = self._config_packages(config["packages"])
        self.default_supported_bugtrackers = [
            tr.bugtracker_type for tr in get_all_subclasses(BugTrackerConfig)
        ] or []

    def _add_bugtrackers_to_pgm(self, cfg_program):
        """
        Link an existing bugtracker to the given program
        """
        exit_config = False

        bts = [
            bt for bt in self.bugtrackers if bt not in cfg_program.bugtrackers
        ]
        while not exit_config and bts:
            for count, bt in enumerate(bts):
                logger.info("{}/ {}".format(count + 1, bt.name))
            bt_idx = read_input(
                Fore.BLUE
                + "Which ones to add for this program ('1,2,3') -  empty to pass: "
                + Style.RESET_ALL
            )
            if bt_idx:
                try:
                    bt_idx = [int(i) for i in bt_idx.split(",")]
                    [
                        cfg_program.bugtrackers.append(bts[i - 1])
                        for i in bt_idx
                    ]
                except Exception as e:
                    logger.error("Invalid Response")
                    continue
            exit_config = True

    def _config_to_keep(self, cfg_program):
        """
        Select bugtrackers to keep in program configuration
        """
        cfg_bt = cfg_program.bugtrackers
        for count, bt in enumerate(cfg_bt):
            logger.info(
                str(count + 1)
                + "/ "
                + Fore.GREEN
                + cfg_program.slug
                + Style.RESET_ALL
                + " tracked on "
                + Fore.GREEN
                + bt.name
                + Style.RESET_ALL
                + " (id: "
                + Fore.GREEN
                + bt.project
                + Style.RESET_ALL
                + ")"
            )
        read_config_to_keep = read_input(
            Fore.BLUE
            + "which ones to keep (eg 1 2 3, empty to keep all programs, None to keep any): "
            + Style.RESET_ALL
        )

        config_to_keep = []
        if read_config_to_keep == "":
            config_to_keep = [config for config in range(1, len(cfg_bt) + 1)]
        elif read_config_to_keep.lower() == "none":
            self.yeswehack = []
            self.bugtrackers = []
            return
        else:
            read_config_to_keep = read_config_to_keep.split(" ")
        invalid_response = False
        for config in read_config_to_keep:
            try:
                if not 0 < int(config) <= len(cfg_bt):
                    invalid_response = True
                    logger.error("Value not in range")
                else:
                    config_to_keep.append(int(config))
            except:
                invalid_response = True
        if invalid_response:
            self._update_configuration()
        return config_to_keep

    def _config_bugtrackers(self, bgtrackers, configure_mode=False):
        """
        Build BugTrackerConfig-s Object from config information.

        :param dict bgtrackers: Bugtrackers definitions.
        :param bool configure_mode: True if configuration mode is actif.
        """
        bts = []
        for bt_name, bt_config in bgtrackers.items():
            if "type" in bt_config:
                class_ = self.get_bugtracker_class(bt_config["type"])
                bts.append(
                    class_(
                        bt_name,
                        configure_mode=configure_mode,
                        no_interactive=self.no_interactive,
                        **bt_config
                    )
                )
            else:
                logger.critical(
                    "type not set in {} configuration part".format(bt_name)
                )
                sys.exit(120)
        return bts

    def _config_packages(self, packages):
        """
        Build ExtraPackageConfig-s Object from config information.

        :param list packages: packages definitions.
        """
        pkgs = []
        for params in packages:
            pkgs.append(ExtraPackageConfig(configure_mode=False, **params))
        return pkgs

    def _config_ywh(self, ywh, configure_mode=False):
        """
        Build YesWeHackConfig-s Object from config information.

        :param dict ywh: yeswehack definitions.
        """
        bts = []
        for ywh_name, ywh_config in ywh.items():
            bts.append(
                YesWeHackConfig(
                    ywh_name,
                    self.bugtrackers,
                    configure_mode=configure_mode,
                    no_interactive=self.no_interactive,
                    **ywh_config
                )
            )
        return bts

    def _load(self, filename):
        """
        Load yaml file with name  `filename`.

        :param str filename: name of file to load.
        """
        if not filename.exists() and not self.configure_mode:
            logger.critical("File {} not exist".format(filename))
            sys.exit(130)
        configuration = {}
        if filename.exists():
            with open(filename, "r") as ymlfile:
                configuration = yaml.safe_load(ymlfile)
        return configuration

    def _update_package(self):
        """
        In configure mode only, used to change configured package.
        """
        for package in self.packages:
            exit_config = False
            package.info()
            while not exit_config:
                del_modules = read_input(
                    Fore.BLUE
                    + "Which modules to delete ? ('1,2,3') - empty to pass :"
                    + Style.RESET_ALL
                )
                if del_modules:
                    try:
                        err = ""
                        del_idx = []
                        for i in del_modules.split(","):
                            idx = int(i) - 1
                            if i not in range(len(package._modules)):
                                err = "Index out of range"
                                break
                            if i in del_idx:
                                continue
                            del_idx.append(idx)
                        if not err:
                            exit_config = True
                        else:
                            logger.error(err)
                            continue
                    except:
                        logger.error("Invalid Input")
                        continue
                exit_config = True

            for i in sorted(del_modules, reverse=True):
                del package.modules[i]

    def _update_configuration(self):
        """
        In configure mode only, used to change configured object.
        """
        self._update_package()

        for cfg_ywh in self.yeswehack:
            for cfg_program in cfg_ywh.programs:
                config_to_keep = self._config_to_keep(cfg_program)

                # Replace current config with the item we keep
                if len(config_to_keep) < len(cfg_program.bugtrackers):
                    cfg_program._delete_bugtrackers(
                        config_to_keep, self.bugtrackers
                    )
                cfg_program.bugtrackers = [
                    cfg_program.bugtrackers[i - 1] for i in config_to_keep
                ]
                self._add_bugtrackers_to_pgm(cfg_program)

    ############################################################
    #################### Static methods ########################
    ############################################################
    @staticmethod
    def get_bugtracker_class(tracker_type):
        """
        Return bugtracker class from tracker type.

        :param str tracker_type: type identifier of the BugTrackerConfig subclass.
        """
        tracker_config = get_all_subclasses(
            BugTrackerConfig, cls_attr_filter={"bugtracker_type": tracker_type}
        )
        if len(tracker_config) != 1:
            if len(tracker_config) == 0:
                raise BugTrackerNotFound(
                    "Bug tracker not found from type `{}`".format(tracker_type)
                )
            else:
                raise MultipleBugTrackerMacth(
                    "Multiple Bug tracker have the same `bugtracker_type` class attribute (bugtracker_type : {}, classes : {}). Please change refered BugTracker in config file.".format(
                        tracker_type,
                        ", ".join(
                            [tr.bugtracker_type for tr in tracker_config]
                        ),
                        tracker_config,
                    )
                )
        return tracker_config[0]



from abc import abstractmethod
from ywh2bt.logging import logger
from colorama import Fore, Style

import sys
import copy
from colorama import Fore, Style

from ywh2bt.utils import read_input
from abc import abstractmethod

__all__ = ["ConfigObject", "BugTrackerConfig"]


"""
Base Configuration classes.
"""


class ConfigObject(object):

    ############################################################
    ######################## Constructor #######################
    ############################################################
    """
    Base class for all configuration class.
    """

    def __init__(self, keys, type_name, **config):
        self.validate(keys, type_name, config)

    ############################################################
    ###################### Static method #######################
    ############################################################

    @staticmethod
    def validate(keys, type_name, config):
        """
        Validate mandatory keys send, and exit if error detected.

        :param list keys: list of key needed in config.
        :param str type_name: keyword identifier of class caller for logging error.
        :param dict config: config dictionary loaded from config yaml file.
        """
        errors = []
        for key in keys:
            if key not in config:
                errors.append(key)
        if errors:
            logger.critical(
                "{} not set in {} configuration part".format(
                    ", ".join(errors), type_name
                )
            )
            sys.exit(120)

    ############################################################
    ####################### Class method #######################
    ############################################################
    @classmethod
    def check_secret_keys(cls, no_interactive, secret_keys, config):
        """
        Validate secret keys in config and log warning if error detected.

        :param bool no_interactive: is it execute in no_interactive mode ?
        :param list secret_keys: keys needed in config if secret is active.
        :param dict config: config dictionary loaded from config yaml file.
        """
        secrect_in_config = []
        for s in secret_keys:
            if s in config:
                secrect_in_config.append(s)
        if secrect_in_config and not no_interactive:
            logger.warning(
                "{} secrets key-s in configuration but non interactive is not actif ".format(
                    ", ".join(secrect_in_config)
                )
            )


class BugTrackerConfig(ConfigObject):

    """
    Loader of a bugtracker config.


    :attr str bugtracker_type: name identifier of the current bugtracker config class
    :attr dict _description: description of given keys.
    :attr list mandatory_keys: keys needed in configuration file.
    :attr list secret_keys: keys define as secret (asked in interactive mode).
    :attr dict optional_keys: keys with default value.
    """

    ############################################################
    ######################## Attributes ########################
    ############################################################
    bugtracker_type = "abstract"
    _description = dict()
    mandatory_keys = []
    secret_keys = []
    optional_keys = dict()
    conditional_keys = dict()
    ############################################################
    ######################## Constructor #######################
    ############################################################
    def __new__(
        cls, name, no_interactive=False, configure_mode=False, **config
    ):
        secrect_in_config = []
        cls.check_secret_keys(no_interactive, cls.secret_keys, config)
        cls._set_properties()
        inst = super().__new__(cls)
        return inst

    def __init__(
        self, name, no_interactive=False, configure_mode=False, **config
    ):
        self._bugtracker = None
        keys = []
        if config or not configure_mode:
            keys += self.mandatory_keys
            if no_interactive:
                keys += self.secret_keys
        type_ = config.get("type", self.bugtracker_type)
        if not "project" in [
            *self.mandatory_keys,
            *self.secret_keys,
            *self.optional_keys_list(),
        ]:
            logger.critical(
                "'project' key not present in mandatory_keys, secret_keys or optional_keys for {} class".format(
                    self.__class__
                )
            )
            sys.exit(210)
        super().__init__(keys, "{} BugTracker".format(type_.title()), **config)
        self._name = name
        self._type = type_
        self._no_interactive = no_interactive
        self._configure_mode = configure_mode

        if config or not configure_mode:
            self._configure_attributes(**config)

        if configure_mode:
            self.configure()

        if not no_interactive:
            self.get_interactive_info()

        if not self._bugtracker:
            self._set_bugtracker()

    ############################################################
    ##################### Instance methods #####################
    ############################################################

    def configure(self, test=True):
        """
        Main methods to configure the implement bugtracker config class.

        :param bool test: if True, test client connexion.
        """
        self.read_mandatory()
        self.read_optional()
        if self.no_interactive:
            self.read_secret()
        self.read_conditional(interactive=self.no_interactive, configure=True)
        self.user_config()
        if self.no_interactive:
            self._set_bugtracker()
            if test:
                self.test_project()

    def get_interactive_info(self):
        """
        Get secret info as interactive.
        """
        self.read_secret()
        self.read_conditional(interactive=True, configure=False)

    def read_mandatory(self):
        """
        Ask user to write values for each mandatories keys.
        """
        for key in self.mandatory_keys:
            desc = ""
            if key in self._description:
                desc += (
                    Fore.GREEN + f" - ({self._description[key]})" + Fore.RESET
                )
            setattr(
                self,
                "_" + key,
                read_input(
                    Fore.BLUE
                    + "{}{}: ".format(key.title(), desc)
                    + Style.RESET_ALL
                ),
            )

    def read_conditional(self, interactive=False, configure=False):
        """
        Ask user to write values for each conditional keys if condition is True.
        """
        for key, info in self.conditional_keys.items():
            desc = ""
            if key in self._description:
                desc += (
                    Fore.GREEN + f" - ({self._description[key]})" + Fore.RESET
                )
            default = info.get("default", None)
            secret = info.get("secret", False)
            condition = info["condition"]
            if condition(self):
                if secret and interactive:
                    setattr(
                        self,
                        "_" + key,
                        read_input(
                            Fore.GREEN
                            + self.type.title()
                            + Fore.BLUE
                            + " {}{}{}: ".format(
                                "{}".format(key.title())
                                + (
                                    " for {}".format(
                                        Fore.GREEN + self.login + Fore.BLUE
                                    )
                                    if hasattr(self, "login")
                                    else ""
                                )
                                + (
                                    " on {}".format(
                                        Fore.GREEN + self.url + Fore.BLUE
                                    )
                                    if hasattr(self, "url")
                                    else ""
                                ),
                                " [default='{}']"
                                if default is not None
                                else "",
                                desc,
                            )
                            + Style.RESET_ALL
                            or default,
                            secret=True,
                        ),
                    )
                elif not secret and default is not None and configure:
                    setattr(
                        self,
                        "_" + key,
                        read_input(
                            Fore.GREEN
                            + self.type.title()
                            + Fore.BLUE
                            + " {} [default='{}']{}: ".format(
                                key.title(), default, desc
                            )
                            + Style.RESET_ALL
                        )
                        or default,
                    )
                elif not secret and default is None and configure:
                    setattr(
                        self,
                        "_" + key,
                        read_input(
                            Fore.BLUE
                            + "{}{}: ".format(key.title(), desc)
                            + Style.RESET_ALL
                        ),
                    )

    def read_optional(self):
        """
        Ask user to write values for each optinals keys.
        """
        for key, default in self.optional_keys.items():
            desc = ""
            if key in self._description:
                desc += (
                    Fore.GREEN + f" - ({self._description[key]})" + Fore.RESET
                )
            setattr(
                self,
                "_" + key,
                read_input(
                    Fore.GREEN
                    + self.type.title()
                    + Fore.BLUE
                    + " {} [default='{}']{}: ".format(
                        key.title(), default, desc
                    )
                    + Style.RESET_ALL
                )
                or default,
            )

    def read_secret(self):
        """
        Ask user to write values for each secrets keys.
        """
        for key in self.secret_keys:
            desc = ""
            if key in self._description:
                desc += (
                    Fore.GREEN + f" - ({self._description[key]})" + Fore.RESET
                )
            setattr(
                self,
                "_" + key,
                read_input(
                    Fore.BLUE
                    + "{}{}: ".format(
                        "{}".format(key.title())
                        + (
                            " for {}".format(
                                Fore.GREEN + self.login + Fore.BLUE
                            )
                            if hasattr(self, "login")
                            else ""
                        )
                        + (
                            " on {}".format(Fore.GREEN + self.url + Fore.BLUE)
                            if hasattr(self, "url")
                            else ""
                        ),
                        desc,
                    )
                    + Style.RESET_ALL,
                    secret=True,
                ),
            )

    def test_project(self):
        """
        Test connexion to the project on the client
        """
        if self.no_interactive:
            logger.info(
                "Testing project: {}{}{}".format(
                    Fore.BLUE, self.project, Style.RESET_ALL
                )
            )
        try:
            self._bugtracker.get_project()
            if self.no_interactive:
                logger.info(
                    "{} status : {}OK{}".format(
                        self.project, Fore.GREEN, Style.RESET_ALL
                    )
                )
            else:
                logger.info(
                    "{project} ok on {url}".format(
                        project=self.project, url=self.url
                    )
                )
        except Exception as e:
            if self.no_interactive:
                logger.info(
                    "{} status : {}KO{}".format(
                        self.project, Fore.RED, Style.RESET_ALL
                    )
                )
                self.configure(test=False)
            else:
                logger.error(
                    "{project} ko on {url}".format(
                        project=self.project, url=self.url
                    )
                )
                sys.exit(-210)

    def to_dict(self):
        """
        Map object to dictionary
        """
        component = {"type": self.type}
        for attr in [*self.mandatory_keys, *self.optional_keys_list()]:
            component[attr] = self.__getattribute__(attr)

        for attr, info in self.conditional_keys.items():
            condition = info["condition"]
            secret = info.get("secret", False)
            default = info.get("default", None)
            if (secret and self.no_interactive) or not secret:
                component[attr] = self.__getattribute__(attr)

        if self.no_interactive:
            for attr in self.secret_keys:
                component[attr] = self.__getattribute__(attr)
        return {self.name: component}

    def user_config(self):
        """
        Methods to implement for special user interaction in configuration mode.
        """
        pass

    def verify(self):
        """
        Test project connexion
        """
        self.test_project()

    def _configure_attributes(self, **config):
        """
            Append protected attribute for each keys in mandatory keys and optional_keys, secret keys if no_interactive,
            in aggreement with config file.
            """
        for attr in self.mandatory_keys:
            setattr(self, "_" + attr, config[attr])
        for attr in self.secret_keys:
            setattr(
                self, "_" + attr, config[attr] if self._no_interactive else ""
            )
        for attr, default in self.optional_keys.items():
            setattr(self, "_" + attr, config.get(attr, default))
        for attr, info in self.conditional_keys.items():
            default = info.get("default", None)
            secret = info.get("secret", False)
            condition = info["condition"]
            if condition(self):
                if secret:
                    if default is not None:
                        setattr(
                            self,
                            "_" + attr,
                            config.get(attr, default)
                            if self._no_interactive
                            else "",
                        )
                    else:
                        setattr(
                            self,
                            "_" + attr,
                            config[attr] if self._no_interactive else "",
                        )
                elif (
                    default is not None
                ):  # secret == False, default is not None
                    setattr(self, "_" + attr, config.get(attr, default))
                else:  #  secret == False, default is None => mandatory
                    setattr(self, "_" + attr, config[attr])

    def _get_bugtracker(self, *args, **kwargs):
        """
        Configure bugtracker client and test if is reachable.
        Args and Kwargs are parameters for bugtracker client class.
        """
        try:
            self._bugtracker = self.client(*args, **kwargs)
            if self.no_interactive:
                logger.info(
                    "{} ({}) status: {}".format(
                        self.bugtracker_type,
                        self.name,
                        "{}OK{}".format(Fore.GREEN, Style.RESET_ALL),
                    )
                )
            else:
                logger.info(
                    "Login ok on {type} ({url})".format(
                        type=self.type, url=self.url
                    )
                )
        except Exception as e:
            if self.no_interactive:
                logger.info(
                    "{} ({}) status: {}".format(
                        self.bugtracker_type,
                        self.name,
                        "{}KO{}".format(Fore.RED, Style.RESET_ALL),
                    )
                )
            else:
                logger.error(
                    "Login fail on {type} ({url}) with error {error}".format(
                        type=self.type, url=self.url, error=str(e)
                    )
                )
                raise e
                sys.exit(-200)
            self.configure()

    def set_yeswehack_domain(self, ywh_domain):
        self._bugtracker.set_yeswehack_domain(ywh_domain)

    ############################################################
    ####################### Class methods ######################
    ############################################################
    @classmethod
    def conditional_keys_list(cls):
        """
        Map conditional_keys Keys to list object.
        """
        return list(cls.conditional_keys.keys())

    @classmethod
    def optional_keys_list(cls):
        """
        Map optional_keys Keys to list object.
        """
        return list(cls.optional_keys.keys())

    @classmethod
    def _set_properties(cls):
        """
        Set get properties for each keys define in mandatory, optonial and secret keys.
        """

        def set_item(cls, attr):
            def func(self):
                return self.__getattribute__("_" + attr)

            setattr(cls, attr, property(copy.copy(func)))

        for attr in [
            *cls.mandatory_keys,
            *cls.secret_keys,
            *cls.optional_keys_list(),
            *cls.conditional_keys_list(),
        ]:
            set_item(cls, attr)

    ############################################################
    ##################### Abstract methods #####################
    ############################################################

    @abstractmethod
    def _set_bugtracker(self):
        """
        Method to setup bugtracker class.
        """
        pass

    ############################################################
    ######################## Properties ########################
    ############################################################
    @property
    def no_interactive(self):
        return self._no_interactive

    @property
    def bugtracker(self):
        return self._bugtracker

    @property
    def configure_mode(self):
        return self._configure_mode

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def bugtracker(self):
        return self._bugtracker



import getpass
from .html2jira import html2jira

__all__ = ["read_input", "get_all_subclasses", "html2jira", "unescape_text"]


"""
Utils Module, define simply fonction.
"""


def unescape_text(text):
    for tag, value in {"&amp;": "&", "&lt;": "<", "&gt;": ">"}.items():
        text = text.replace(tag, value)
    return text


def read_input(text, secret=False):
    """
    Wrapper for input and getpass with color print text.
    :param text: text to  show before read input.
    :param bool secret: User input is invisible or not.
    """

    print(text, flush=True, end="")
    if secret:
        return getpass.getpass(prompt="")
    return input()


def get_all_subclasses(cls, ret=None, cls_attr_filter={}):
    """
    get subclasses from an ancestor recurively.
    You can use on attribute value filter.

    :param class cls: input class
    :param list ret: return list
    :param dict cls_attr_filter: filter for subclasse attribute on value.
    """
    for sub in cls.__subclasses__():
        keep = True
        for key, item in cls_attr_filter.items():
            if sub.__dict__.get(key, None) != item:
                keep = False
                break
        if keep:
            if ret is None or not isinstance(ret, list):
                ret = []
            ret.append(sub)
        get_all_subclasses(sub, ret=ret, cls_attr_filter=cls_attr_filter)
    return ret if ret is not None else []



from .abstract import ConfigObject
from ywh2bt.logging import logger
from colorama import Fore, Style

from yeswehack.api import YesWeHack
from yeswehack.exceptions import (
    BadCredentials,
    ObjectNotFound,
    InvalidResponse,
    TOTPLoginEnabled,
)
from ywh2bt.utils import read_input
import sys

__all__ = ["YesWeHackConfig", "ProgramConfig"]

"""
YesWeHack Config module
"""


class YesWeHackConfig(ConfigObject):

    """
    Build/Configure YesWeHack Configuration.

    :attr str default_url_api: default URL Api for this application.
    :attr list know_header: headers known to exchange with api.
    """

    default_url_api = "https://apps.yeswehack.com"
    known_headers = ["X-YesWeHack-Apps"]

    ############################################################
    ###################### Constructor #########################
    ############################################################
    def __init__(
        self,
        name,
        bugtrackers,
        no_interactive=False,
        configure_mode=False,
        **config,
    ):
        assert bugtrackers is not None
        self._name = name
        self.configure_mode = configure_mode
        self.no_interactive = no_interactive
        self._programs = []
        self._oauth_args = {}
        self._apps_headers = {}
        self._password = ""
        self._totp_secret = ""

        self.check_secret_keys(
            no_interactive, ["totp_secret", "password"], config
        )

        if config or not configure_mode:
            self._totp = config.get("totp", False)
            keys = ["login", "programs"]
            if self.no_interactive:
                keys.append("password")
                if self._totp:
                    keys.append("totp_secret")
            super().__init__(keys, "YesWeHack", **config)
            self._login = config["login"]
            self._api_url = config.get("api_url", self.default_url_api)
            self._totp_secret = (
                config["totp_secrect"]
                if self.no_interactive and self._totp
                else ""
            )
            self._oauth_mode = "oauth_args" in config
            self._oauth_args = config.get("oauth_args", {})
            self._apps_headers = config.get("apps_headers", {})
            self._totp = config.get("totp", False)
            self._programs = self._config_programs(
                bugtrackers, config["programs"]
            )

            if not no_interactive and not configure_mode:
                self.get_interactive_info()
            else:
                self._password = (
                    config["password"] if self.no_interactive else ""
                )

            self._ywh = YesWeHack(
                username=self._login,
                password=self._password,
                api_url=self._api_url,
                oauth_mode=self._oauth_mode,
                oauth_args=self._oauth_args,
                apps_headers=self.apps_headers,
                headers={"Access-Control-Allow-Originx": "*"},
                lazy=False,
            )
        if configure_mode:
            self.configure(bugtrackers)

    ############################################################
    ################### Instance Methods #######################
    ############################################################

    def configure(self, bugtrackers):
        """
        Configure programs interactively.
        """
        self.config_user()
        exit_config = False
        while not exit_config:
            self._programs.append(
                ProgramConfig(
                    bugtrackers,
                    no_interactive=self.no_interactive,
                    configure_mode=True,
                )
            )

            if read_input(
                Fore.BLUE
                + "Configure an other Program ? [y/N]"
                + Style.RESET_ALL
            ) in ["n", "N", ""]:
                exit_config = True

    def config_oauth(self):
        """
        Configure oauth public information interactively.
        """
        self._oauth_args["client_id"] = read_input(
            Fore.BLUE + "Client id: " + Style.RESET_ALL
        )
        self._oauth_args["redirect_uri"] = read_input(
            Fore.BLUE + "Redirect URI: " + Style.RESET_ALL
        )

    def config_secret(self):
        """
        Configure secret information interactively.
        """
        self._password = read_input(
            Fore.BLUE
            + "Password for {} on {}: ".format(
                Fore.GREEN + self._login + Fore.BLUE,
                Fore.GREEN + self._api_url + Fore.BLUE,
            )
            + Style.RESET_ALL,
            secret=True,
        )
        if self.oauth_mode:
            self._oauth_args["client_secret"] = read_input(
                Fore.BLUE + "Client Secret: " + Style.RESET_ALL, secret=True
            )
        if self.totp:
            self._totp_secret = read_input(
                Fore.BLUE + "Totp secret: " + Style.RESET_ALL, secret=True
            )
        self._config_header()

        self._ywh = YesWeHack(
            username=self._login,
            password=self._password,
            api_url=self._api_url,
            oauth_mode=self._oauth_mode,
            oauth_args=self._oauth_args,
            apps_headers=self.apps_headers,
        )
        self.test_login_config()

    def config_user(self):
        """
        Configure user login information interactively.
        """
        self._api_url = (
            read_input(
                "{}API url [{}{}{}]: {}".format(
                    Fore.BLUE,
                    Fore.GREEN,
                    self.default_url_api,
                    Fore.BLUE,
                    Style.RESET_ALL,
                )
            )
            or self.default_url_api
        )
        self._login = read_input(Fore.BLUE + "YesWeHack login: " + Fore.RESET)
        totp = read_input(
            Fore.BLUE + "Is TOTP Enabled : [y/N] " + Style.RESET_ALL
        )
        self._totp = True if totp in ["y", "Y"] else False

        while True:
            inpt = read_input(
                Fore.BLUE + "OAuth2 Authentication ? [y/N]: " + Fore.RESET
            )
            if inpt in ["y", "Y"]:
                self._oauth_mode = True
                self.config_oauth()
                break
            elif inpt in ["", "n", "N"]:
                self._oauth_mode = False
                break

        if self.no_interactive:
            self.config_secret()

    def get_interactive_info(self):
        """
        Configure secret information interactively.
        """
        logger.info(
            "Getting account info for "
            + Fore.GREEN
            + ", ".join([pgm.name for pgm in self.programs])
            + Style.RESET_ALL
            + " on "
            + Fore.GREEN
            + self.api_url
            + Style.RESET_ALL
            + " via "
            + Fore.GREEN
            + self.login
            + Style.RESET_ALL
        )
        self.config_secret()

    def get_totp_code(self):
        """
        Get TOTP code
        """
        totp_code = None
        if self._totp_secret:
            totp = pyotp.TOTP(self.totp_secret)
            totp_code = totp.now()
        return totp_code

    def test_login_config(self):
        """
        Test login configuration information.
        """
        if self.no_interactive:
            logger.info("Testing login")
        try:
            totp_code = self.get_totp_code()
            self.ywh.login(totp_code=totp_code)
            if self.no_interactive:
                logger.info(
                    "YesWeHack status: " + Fore.GREEN + "OK" + Style.RESET_ALL
                )
            else:
                logger.info(
                    "Login ok with {login} on {url}".format(
                        login=self.login, url=self.api_url
                    )
                )
        except TOTPLoginEnabled:
            if self.no_interactive:
                logger.info(Fore.RED + "KO (TOTP)" + Style.RESET_ALL)
                self._totp = True
                self.config_user()
        except BadCredentials:
            if self.no_interactive:
                logger.info(Fore.RED + "KO" + Style.RESET_ALL)
                self.config_user()
            else:
                logger.error(
                    "Login fail with {login} on {url} (BadCredentials)".format(
                        login=self.login, url=self.api_url
                    )
                )
                sys.exit(100)
        except InvalidResponse:
            if self.no_interactive:
                logger.info(Fore.RED + "KO" + Style.RESET_ALL)
                self.config_user()
            else:
                logger.error(
                    "Login fail with {login} on {url} (InvalidResponse)".format(
                        login=self.login, url=self.api_url
                    )
                )
                sys.exit(100)

    def test_program_config(self):
        """
        Test programs configuration information.
        """
        totp_code = self.get_totp_code()
        self.ywh.login(totp_code=totp_code)
        managed_programs = set(self.ywh.managed_programs())
        pgms_names = [pgm.name for pgm in self.programs]
        diffs = set(pgms_names).difference(managed_programs)
        if diffs:
            logger.error(
                "Program-s {} are not manageable for {} yeswehack user".format(
                    "".join(diffs), self.login
                )
            )
            sys.exit(110)
        if self.no_interactive:
            logger.info("Testing program")
        for program in self.programs:
            try:
                pgm = self.ywh.get_program(program)
                if self.no_interactive:
                    logger.info(
                        "{} status: ".format(self._project)
                        + Fore.GREEN
                        + "OK"
                        + Style.RESET_ALL
                    )
                else:
                    logger.info(
                        "Program {program} ok with {login} on  {url}".format(
                            program=pgm.name,
                            login=self.login,
                            url=self.api_url,
                        )
                    )
            except ObjectNotFound:
                if self.no_interactive:
                    logger.info(
                        "{} status: ".format(self._project)
                        + Fore.RED
                        + "KO"
                        + Style.RESET_ALL
                    )
                else:
                    logger.error(
                        "Program {program} fail with {login} on  {url}".format(
                            program=pgm.name,
                            login=self.login,
                            url=self.api_url,
                        )
                    )
                    sys.exit(110)
                self.config_program_info()

    def to_dict(self):
        """
        Map object to dictionary
        """
        pgms = [pgm.to_dict() for pgm in self.programs]

        component = {
            "api_url": self.api_url,
            "login": self.login,
            "totp": self.totp,
            "programs": pgms,
        }

        if self.oauth_mode:
            component["oauth_args"] = self.oauth_args
        if self.apps_headers:
            component["apps_headers"] = self.apps_headers
        if self.no_interactive:
            component["password"] = self.password
            if self.totp:
                component["totp_secret"] = self.totp_secret

        return {self.name: component}

    def verify(self):
        """
        Verify connexion information
        """
        self.test_login_config()
        self.test_program_config()

    def _config_programs(self, bugtrackers, config, configure_mode=False):
        """
        Configure existing programs.
        """
        pgms = []
        for pgm_config in config:
            pgms.append(
                ProgramConfig(
                    bugtrackers,
                    no_interactive=self.no_interactive,
                    configure_mode=configure_mode,
                    **pgm_config,
                )
            )
        return pgms

    def _config_header(self):
        """
        Config headers interactively
        """
        logger.info("Yeswehack Apps Headers configuration")
        for know_header in self.known_headers:
            header = read_input(
                Fore.BLUE
                + "Value for {}: ".format(Fore.GREEN + know_header + Fore.BLUE)
                + Fore.RESET,
                secret=True,
            )
            if header:
                self.apps_headers[know_header] = header
        inpt = ""
        exit_config = False
        while not exit_config:
            while True:
                inpt = read_input(
                    Fore.BLUE
                    + "Append an other Application Header ? [y/N]: "
                    + Fore.RESET
                )
                if inpt in ["y", "Y"]:
                    apps_header_name = read_input(
                        Fore.BLUE + "Header Name: " + Fore.RESET
                    )
                    apps_header_value = read_input(
                        Fore.BLUE + "Header Value: " + Fore.RESET
                    )
                    self.apps_headers[apps_header_name] = apps_header_value
                    break
                if inpt in ["", "n", "N"]:
                    exit_config = True
                    break

    ############################################################
    ####################### Properties #########################
    ############################################################

    @property
    def apps_headers(self):
        return self._apps_headers

    @property
    def oauth_args(self):
        return self._oauth_args

    @property
    def oauth_mode(self):
        return self._oauth_mode

    @property
    def name(self):
        return self._name

    @property
    def totp_secret(self):
        return self._totp_secret

    @property
    def api_url(self):
        return self._api_url

    @property
    def programs(self):
        return self._programs

    @property
    def login(self):
        return self._login

    @property
    def password(self):
        return self._password

    @property
    def totp(self):
        return self._totp

    @property
    def ywh(self):
        return self._ywh


class ProgramConfig(ConfigObject):

    """
    Load or configure Program configuration.
    """

    ############################################################
    ###################### Constructor #########################
    ############################################################

    def __init__(
        self, bugtrackers, no_interactive=False, configure_mode=False, **config
    ):  # name, bugtrackers={}):
        self.bugtrackers = []
        if config or not configure_mode:
            super().__init__(["slug", "bugtrackers_name"], "Program", **config)
            self._slug = config["slug"]
            self._validate_and_set_bugtrackers(
                config["bugtrackers_name"], bugtrackers
            )

        if configure_mode:
            self.configure(bugtrackers)

    ############################################################
    ################### Instance Methods #######################
    ############################################################

    def configure(self, bugtrackers):
        """
        Configure and link to bugtrackers interactively.
        """
        all_bt_names = [bt.name for bt in bugtrackers]
        self._slug = read_input(Fore.BLUE + "Program slug: " + Style.RESET_ALL)
        exit_config = False
        while not exit_config:
            bg_list = [
                "{}/ {}".format(i + 1, j)
                for i, j in enumerate(all_bt_names)
                if not j in [bt.name for bt in self.bugtrackers]
            ]
            if not bg_list:
                exit_config = True
                continue
            try:
                bt_idx = read_input(
                    Fore.BLUE
                    + "Select a bugtracker :{} \n\t{}\n".format(
                        Style.RESET_ALL, "\n\t".join(bg_list)
                    )
                )
                self.bugtrackers.append(bugtrackers[int(bt_idx) - 1])
                if read_input(
                    Fore.BLUE
                    + "Add an other bugtracker ? [y/N]"
                    + Style.RESET_ALL
                ) in ["n", "N", ""]:
                    exit_config = True
            except Exception as e:
                logger.error("Invalide response")

    def to_dict(self):
        """
        Map object to dictionary.
        """
        return {
            "slug": self.slug,
            "bugtrackers_name": [
                bgtracker.name for bgtracker in self.bugtrackers
            ],
        }

    def _delete_bugtrackers(self, config_to_keep, bugtrackers):
        """
        Delete Linked bugtrackers
        """
        delete_bugtrackers = []

        while True:
            user_input = read_input(
                Fore.BLUE
                + "You Want to delete other bugtrackers ? "
                + Fore.YELLOW
                + "(if not, they are just unset for this program, otherwise, this operation could break you're configuration file)"
                + Fore.BLUE
                + " [y/N] :"
                + Style.RESET_ALL
            )
            if user_input in ["n", "N", ""]:
                break
            elif user_input in ["y", "Y"]:
                delete_bugtrackers = [
                    bugtrackers[i - 1]
                    for i in range(1, len(bugtrackers) + 1)
                    if i not in config_to_keep
                ]
                break
        for del_bg in delete_bugtrackers:
            try:
                bugtrackers.remove(del_bg)
            except:
                logger.warning(
                    "You can't delete unexisting BugTracker ({})".format(
                        del_bg.name
                    )
                )

    def _validate_and_set_bugtrackers(self, program_bt_names, bugtrackers):
        """
        Configure and link to bugtrackers from configuration file.
        """
        all_bt_names = [bt.name for bt in bugtrackers]
        warnings = []
        for bt in program_bt_names:
            if bt not in all_bt_names:
                warnings.append(bt)
            else:
                self.bugtrackers.append(bugtrackers[all_bt_names.index(bt)])
        if not self.bugtrackers:
            logger.critical(
                "Bugtrackers type set in progams config part for {} does'nt exist in bugtrackers global configuration part".format(
                    self.slug
                )
            )
            sys.exit(120)
        if warnings:
            logger.warning(
                "{} doesn't exists in bugtrackers global configuration part".format(
                    ", ".join(warnings)
                )
            )

    ############################################################
    ####################### Properties #########################
    ############################################################
    @property
    def slug(self):
        return self._slug



import logging
import coloredlogs
import os
import urllib3

__all__ = ["_setup_logger", "logger"]
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

"""
Logging Module.
"""


def _setup_logger(level="INFO"):
    """
    Setup a Color Logger.

    :param str level: Logging level.
    """
    logging.getLogger().setLevel("ERROR")

    logger = logging.getLogger(__name__)

    fmt = "[%(levelname)s] %(asctime)s %(message)s"
    datefmt = "%m/%d/%Y %H:%M:%S"

    FIELD_STYLES = dict(
        asctime=dict(color="green"),
        levelname=dict(color="magenta", bold=coloredlogs.CAN_USE_BOLD_FONT),
    )

    LEVEL_STYLES = dict(
        debug=dict(color="green"),
        info=dict(color="cyan"),
        verbose=dict(color="blue"),
        warning=dict(color="yellow"),
        error=dict(color="red"),
        critical=dict(color="red", bold=coloredlogs.CAN_USE_BOLD_FONT),
    )
    coloredlogs.install(
        level=level,
        logger=logger,
        reconfigure=False,
        fmt=fmt,
        datefmt=datefmt,
        level_styles=LEVEL_STYLES,
        field_styles=FIELD_STYLES,
    )
    logger.propagate = False
    return logger


if not locals().get("logger", None):
    logger = _setup_logger()



# -*- encoding: utf-8 -*-

import github
from .bugtracker import BugTracker
from github.GithubException import BadCredentialsException
from ywh2bt.config import BugTrackerConfig
from bs4 import BeautifulSoup
import requests
from requests_toolbelt import MultipartEncoder
from ywh2bt.logging import logger
from ywh2bt.utils import read_input
from colorama import Fore, Style
from copy import copy
import re

__all__ = ["YWHGithub", "YWHGithubConfig"]

"""
Github Bugtracker System
"""


class YWHGithub(BugTracker):

    """
    Github Client Wrapper
    """

    ############################################################
    ####################### Constructor ########################
    ############################################################
    def __init__(
        self, url, project, token, login="", password="", github_cdn_on=False
    ):
        self.url = url
        self.project = project
        self.token = token
        self.password = password
        self.username = login
        self.github_cdn_on = github_cdn_on
        self.session = None
        if self.url.endswith("/"):
            self.url = self.url[:-1]

        if self.url != "https://api.github.com":
            self.bt = github.Github(
                base_url=self.url, login_or_token=self.token
            )
            self.github_domain = (
                self.url.replace("https://", "")
                .replace("http://", "")
                .split("/")[0]
            )
        else:
            self.bt = github.Github(login_or_token=self.token)
            self.github_domain = "github.com"
        try:
            self.bt.get_user().name
        except BadCredentialsException:
            raise

    ############################################################
    #################### Instance methods ######################
    ############################################################
    def get_project(self):
        """
        return github project object
        """
        try:
            repo = self.bt.get_repo(self.project)
        except github.GithubException.UnknownObjectException:
            raise
        return repo

    def post_issue(self, report):
        """
        post issue to github from report information

        :param yeswehack.api.Report report: Report from yeswehack
        """
        copy_report = copy(report)
        repo = self.bt.get_repo(self.project)
        title = self.report_as_title(report)
        body = self.report_as_description(report)
        issue = repo.create_issue(title=title, body=body)
        substitute_regex = r"!?\[[^\[\]]*\]\({text}\)"
        attach_name_regex = "(?:!?\[)([^\[\]]*)(?:\])(?:\({text}\))"
        for attachment in report.attachments:
            attach_name = re.findall(
                attach_name_regex.format(text=re.escape(attachment.url)), body
            )
            if attach_name:
                attach_name = attach_name[0]
            else:
                attach_name = attachment.name
            if self.github_cdn_on:
                attachment.get_data()
                url, status_code = self.post_attachment(
                    attachment, issue.number
                )
                if url:
                    body = body.replace(attachment.url, url)
                else:
                    body = re.sub(
                        substitute_regex.format(
                            text=re.escape(attachment.url)
                        ),
                        '(Attachment "{f_name}" not available due to upload error)'.format(
                            f_name=attach_name
                        ),
                        body,
                    )
            else:
                body = re.sub(
                    substitute_regex.format(text=re.escape(attachment.url)),
                    '(Attachment "{f_name}" not available due to export script’s configuration)'.format(
                        f_name=attach_name
                    ),
                    body,
                )
        issue.edit(body=body)
        return issue

    def get_url(self, issue):
        """
        Return issue url
        """
        return issue.html_url

    def get_id(self, issue):
        """
        Return issue id
        """
        return issue.number

    def cdn_request(self, method, url, data=None, headers={}):
        try:
            if headers:
                requested = self.session.request(
                    method, url, data=data, headers=headers
                )
            else:
                requested = self.session.request(method, url, data=data)
        except Exception as e:
            logger.error(f"An exception occur on {self.name} : {str(e)}")
            raise e
        return requested

    def login(self):
        """
        Login to github platform
        """
        status = None
        if self.session is None:
            self.session = requests.Session()
            r = self.cdn_request("GET", f"https://{self.github_domain}/login")
            form = BeautifulSoup(r.text, features="lxml").find("form")
            keys = {}
            for key in [
                "authenticity_token",
                "commit",
                "timestamp_secret",
                "webauthn-support",
                "webauthn-iuvpaa-support",
                "required_field_59db",
                "utf8",
            ]:
                value = form.find("input", {"name": key})
                keys = {**keys, key: value["value"] if value else None}
            status = self.session.post(
                f"https://{self.github_domain}/session",
                data={
                    **keys,
                    "login": self.username,
                    "password": self.password,
                },
            ).status_code
        return status

    def logout(self):
        """
        Logout from github platform
        """
        status = None
        if self.session is not None:
            r = self.cdn_request("GET", f"https://{self.github_domain}/")
            hiddens = (
                BeautifulSoup(r.text, features="lxml")
                .find("form", {"action": "/logout"})
                .findAll("input", {"type": "hidden"})
            )
            data = {}
            for h in hiddens:
                data = {**data, h["name"]: h["value"]}
            status = self.cdn_request(
                "POST", f"https://{self.github_domain}/logout", data=data
            ).status_code
            self.session = None
        return status

    def post_attachment(self, attachment, issue_id):
        """
        Post attachment to issue on github CDN

        :param yeswehack.api.Attachment attachment: Attachment file representation.
        :param int issue_id: issue_id on project.
        """
        url = ""
        status_code = self.login()
        href = None
        if status_code != 200:
            msg = 'Github CDN ("https://{self.github_domain}") login error'
            logger.error(msg)
            raise Exception(msg)

        if self.session:
            repo_id = self.bt.get_repo(self.project).id

            r = self.cdn_request(
                "GET",
                "https://{}/{}/issues/{}".format(
                    self.github_domain, self.project, issue_id
                ),
            )

            file_attach = BeautifulSoup(r.text, features="lxml").find(
                "file-attachment"
            )

            if not file_attach:
                msg = f"Can't get information to upload data, status code: {r.status_code}"
                logger.error(msg)
                raise Exception(msg)
            filename = attachment.original_name.split("/")[-1]
            content_type = attachment.mime_type
            fields = {
                "name": filename,
                "size": str(len(attachment.data)),
                "content_type": content_type,
                "authenticity_token": (
                    file_attach["data-upload-policy-authenticity-token"]
                ),
                "repository_id": "{}".format(repo_id),
            }
            data = MultipartEncoder(fields=fields)
            href = self.cdn_request(
                "POST",
                f"https://{self.github_domain}/upload/policies/assets",
                data=data,
                headers={
                    **self.session.headers,
                    "Content-Type": data.content_type,
                    "Content-Length": str(data.len),
                },
            )
            status_code = href.status_code
            if status_code == 201:  #  status_code 201
                info = href.json()
                if "asset" not in info:
                    msg = f"Can't upload data, response: {info}"
                    logger.error(msg)
                    raise Exception(msg)
                url = info["asset"]["href"]
                fields = {}
                for field, value in info["form"].items():
                    fields[field] = value
                fields["file"] = (filename, attachment.data, content_type)
                data = MultipartEncoder(fields=fields)
                href = self.cdn_request(
                    "POST",
                    info["upload_url"],
                    data=data,
                    headers={
                        **self.session.headers,
                        "Content-Type": data.content_type,
                        "Content-Length": str(data.len),
                    },
                )
                status_code = href.status_code

                data = MultipartEncoder(
                    fields={
                        "authenticity_token": info[
                            "asset_upload_authenticity_token"
                        ]
                    }
                )
                r = self.cdn_request(
                    "PUT",
                    f"https://{self.github_domain}" + info["asset_upload_url"],
                    data=data,
                    headers={
                        **self.session.headers,
                        "Content-Type": data.content_type,
                        "Content-Length": str(data.len),
                        "Referer": "https://{}/{}/issues/{}".format(
                            self.github_domain, self.project, issue_id
                        ),
                        "Accept": "application/json",
                    },
                )
                status_code = r.status_code
                try:
                    info = r.json()
                    url = info["href"]
                except:
                    url = ""
            self.logout()
            return url, status_code


class YWHGithubConfig(BugTrackerConfig):

    """
    BugTrackerConfig Class associated to github.

    :attr str bugtracker_type: bugtracker identifier
    :attr class client: Client class
    :attr list mandatory_keys: keys needed in configuration file part
    :attr list secret_keys: keys need interaction in interactive mode
    :attr dict optional_keys: keys with default values
    :attr dict _description: descriptor for each specified key
    """

    bugtracker_type = "github"
    client = YWHGithub

    mandatory_keys = ["project"]
    secret_keys = ["token"]
    optional_keys = dict(url="https://api.github.com", github_cdn_on=False)
    _description = dict(
        project="path/to/project",
        github_cdn_on="Enable or Disable saving attachment file to Github CDN",
    )

    conditional_keys = dict(
        password=dict(condition=lambda x: x._github_cdn_on, secret=True),
        login=dict(condition=lambda x: x._github_cdn_on, secret=False),
    )

    ############################################################
    #################### Instance methods ######################
    ############################################################
    def _set_bugtracker(self):
        if self._github_cdn_on:
            self._get_bugtracker(
                self._url,
                self._project,
                self._token,
                login=self._login,
                password=self._password,
                github_cdn_on=self._github_cdn_on,
            )
        else:
            self._get_bugtracker(self._url, self._project, self._token)

    def user_config(self):
        self._github_cdn_on = (
            True
            if self._github_cdn_on.lower().strip() in ["true", "y", "yes", "1"]
            else False
        )



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

import sys
import os
import yaml
import getpass
from pathlib import Path
from colorama import Fore, Style, init, deinit
from optparse import OptionParser
from yeswehack.api import YesWeHack
from yeswehack.exceptions import (
    BadCredentials,
    ObjectNotFound,
    InvalidResponse,
    TOTPLoginEnabled,
)
from ywh2bt.trackers.bugtracker import BugTracker
from ywh2bt import config
from ywh2bt.logging import logger

__all__ = ["main"]
__VERSION__ = "0.2"

"""
Entry point for script and setup
"""


def main():
    """
    Parse args from commande line, setup GlobalConfig and run corresponding process.
    """
    init()

    description = (
        "ywh2bt is a simple YesWeHack tools for Bounty Program manager."
        + "it's build to exchange report log from the platform to your issue manager."
    )

    parser = OptionParser(
        usage="",
        version="%prog {}".format(__VERSION__),
        description=description,
    )
    parser.add_option(
        "-c",
        "--configure",
        action="store_true",
        dest="configure",
        help="Create config file",
        default=False,
    )
    parser.add_option(
        "-f", "--filename", action="store", dest="filename", default=""
    )

    parser.add_option(
        "-n",
        "--no-interactive",
        action="store_true",
        dest="no_interactive",
        default=False,
        help="non interactive mode. Store credencials on disk. You need to activate it on configure to store credencials",
    )

    parser.set_usage(
        "usage: %prod [-f|--filename] FILENAME [-n|--no-interactive] [-c|--configure]\n\n{}".format(
            parser.format_help()
        )
    )
    (options, args) = parser.parse_args()

    ywh_cfg = config.GlobalConfig(
        no_interactive=options.no_interactive,
        filename=options.filename,
        configure_mode=options.configure,
    )

    if not options.configure:
        run(ywh_cfg, options)
    deinit()


def run(cfg, options):
    """
    Process YesWeHack Report Log loading and insert them in bugtrackers systems.
    """
    for cfg_ywh in cfg.yeswehack:
        # Iterate on every referenced program
        logger.info("Get info for " + cfg_ywh.name)
        ywh_domain = ".".join(
            cfg_ywh.api_url.replace("https://", "")
            .replace("http://", "")
            .split(".")[1:]
        )
        ywh_domain = ywh_domain.split("/")[0]
        for cfg_pgm in cfg_ywh.programs:

            reports = cfg_ywh.ywh.get_reports(
                cfg_pgm.slug,
                filters={"filter[trackingStatus][]": "AFI"},
                lazy=True,
            )
            for report in reports:
                report = cfg_ywh.ywh.get_report(report.id)
                logger.info("Checking " + report.title)

                comments = report.get_comments(lazy=True)
                for cfg_bt in cfg_pgm.bugtrackers:
                    try:
                        cfg_bt.set_yeswehack_domain(ywh_domain)
                        issue = cfg_bt.bugtracker.post_issue(report)
                        issue_meta = {
                            "url": cfg_bt.bugtracker.get_url(issue),
                            "id": cfg_bt.bugtracker.get_id(issue),
                        }

                        logger.info(report.title + " posted to " + cfg_bt.name)

                        marker = BugTracker.ywh_comment_marker.format(
                            url=cfg_bt.url, project_id=cfg_bt.project
                        )

                        comment = (
                            marker
                            + "\n"
                            + BugTracker.ywh_comment_template.format(
                                type=cfg_bt.type,
                                issue_id=issue_meta["id"],
                                bug_url=issue_meta["url"],
                            )
                        )
                        resp = report.put_tracking_status(
                            "T",
                            cfg_bt.name,
                            issue_meta["url"],
                            tracker_id=issue_meta["id"],
                            message=comment,
                        )

                        try:
                            resp_json = resp.json()
                        except:
                            logger.error("Response from YesWeHack not JSON")
                        else:
                            if "error" in resp_json:
                                logger.error(
                                    "Status Update Error : {}".format(
                                        resp_json.get(
                                            "error_description", resp.text
                                        )
                                    )
                                )
                            else:
                                logger.info("Status updated.")
                    except:
                        logger.error(
                            "An error occur on {}, continue to the next bugtracker".format(
                                cfg_bt.name
                            )
                        )



import setuptools
import ywh2bt
from ywh2bt.ywh2bt import __VERSION__

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "requests",
    "python-gitlab",
    "pygithub",
    "jira",
    "colorama",
    "pyotp",
    "pyyaml",
    "html2text",
    "coloredLogs",
    "yeswehack",
    "beautifulsoup4",
    "lxml",
]

setuptools.setup(
    name=ywh2bt.name,
    version=__VERSION__,
    author="Jean Lou Hau",
    author_email="jl.hau@yeswehack.com",
    description="YesWeHack BugTracker",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://yeswehack.com",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["ywh-bugtracker=ywh2bt.ywh2bt:main"]},
)



# -*- encoding: utf-8 -*-

from .abstract import ConfigObject, BugTrackerConfig
from .yeswehack import YesWeHackConfig, ProgramConfig
from .main import GlobalConfig
from .packages import ExtraPackageConfig

__all__ = [
    "GlobalConfig",
    "ConfigObject",
    "BugTrackerConfig",
    "YesWeHackConfig",
    "ProgramConfig",
    "ExtraPackageConfig",
]

"""
Configurations classes definition
"""



from vstsclient.vstsclient import VstsClient
from vstsclient.models import JsonPatchDocument, JsonPatchOperation
from vstsclient.constants import SystemFields, MicrosoftFields
from colorama import Fore, Style
from .ywhtfs import YWHTfs, YWHTfsConfig, NewVstsClient
from ywh2bt.utils import read_input


class YWHAzuredevops(YWHTfs):
    def __define_client__(self, *args, **kwargs):
        self.bt = NewVstsClient(
            args[0], personal_access_token=args[1], scheme=kwargs["scheme"]
        )


class YWHAzuredevopsConfig(YWHTfsConfig):
    # base_url = "dev.azure.com"
    bugtracker_type = "azuredevops"
    client = YWHAzuredevops
    mandatory_keys = [
        key for key in YWHTfsConfig.mandatory_keys if key != "url"
    ]
    optional_keys = {**YWHTfsConfig.optional_keys, "url": "dev.azure.com"}

    def user_config(self):

        orga = read_input(
            Fore.BLUE
            + "organisation name if you don't have set it in url '{url}/<orga>':".format(
                url=url
            )
            + Style.RESET_ALL
        )

        # project_types = self.bt.get_workitem_types(self.project_name)
        self._url = "{url}/{orga}".format(url=self.url, orga=orga)



# -*- encoding: utf-8 -*-

import jira
import html2text
from .bugtracker import BugTracker
from ywh2bt.config import BugTrackerConfig
from ywh2bt.utils import html2jira, unescape_text
from bs4 import BeautifulSoup
from pprint import pprint
from copy import copy
from string import Formatter

__all__ = ["YWHJira", "YWHJiraConfig"]


class YWHJira(BugTracker):
    """
    Jira Client Wrapper
    """

    description_template = """
    || Title || {local_id} : {title} ||
    | Bug Type | [{bug_type__name}]({bug_type__link}) => [Remediation]({bug_type__remediation_link}) |
    | Scope | {scope} |
    | Severity | {cvss__criticity}, score {cvss__score:.1f}, vector {cvss__vector}|
    | Endpoint |{end_point}|
    | Vulnerable part | {vulnerable_part} |
    | Part Name | {part_name} |
    | Payload | {payload_sample} |
    | Technical Environment | {technical_information}|

    {description_html}
    """

    tag_format = "yeswehack_code_section_{idx}"
    ############################################################
    ####################### Constructor ########################
    ############################################################
    def __init__(self, url, login, password, project, issuetype="Task"):
        self.url = url
        self.login = login
        self.password = password
        self.project = project
        self.issuetype = issuetype
        try:
            self.jira = jira.JIRA(self.url, auth=(self.login, self.password))
        except jira.exceptions.JIRAError:
            raise

    ############################################################
    #################### Instance methods ######################
    ############################################################
    def get_project(self):
        try:
            self.jira.project(self.project)
        except jira.exceptions.os:
            raise

    def post_issue(self, report):
        copy_report = copy(report)
        html = html2text.HTML2Text()
        html.ignore_links = True
        desc, tags = self.report_as_description(copy_report)
        issue_data = {
            "project": {"key": self.project},
            "summary": self.report_as_title(copy_report),
            "description": desc,
            "issuetype": {"name": self.issuetype},
        }
        try:
            issue = self.jira.create_issue(**issue_data)
        except:  #  if jira need direct issuetype name
            issue_data["issuetype"] = self.issuetype
            issue = self.jira.create_issue(**issue_data)
        for attachment in report.attachments:
            attachment.get_data()
            attach = self.jira.add_attachment(
                issue=issue,
                filename=attachment.original_name,
                attachment=attachment.data,
            )
            replacer = (
                "/".join(attach.content.split("/")[:-1])
                + "/"
                + attach.filename.replace(" ", "%20")
            ).replace("\n", "")
            copy_report.description_html = copy_report.description_html.replace(
                attachment.url, replacer
            )
        desc, _ = self.report_as_description(copy_report, tags=tags)
        issue.update(description=desc.replace("!!", "! !"))
        return issue

    def get_url(self, issue):
        return issue.permalink()

    def get_id(self, issue):
        return issue.key

    def report_as_description(
        self, report, template=None, additional_keys=[], tags=[]
    ):
        report.description_html = self.replace_external_link(report)
        report.description_html = self._img_to_jira_tag(
            report.description_html
        )

        report.description_html, tmp_tags = self._code_to_jira_tag(
            report.description_html
        )
        if not tags:
            tags = tmp_tags
        description = super().report_as_description(
            report,
            template=template,
            additional_keys=additional_keys,
            html_parser=html2jira,
        )

        for (_, f_name, _, _) in Formatter().parse(description):
            if f_name in tags:
                description = description.replace(
                    "{%s}" % f_name, tags[f_name]
                )
        return description, tags

    def _code_to_jira_tag(self, html):
        soup = BeautifulSoup(html, features="lxml")
        n_html = str(soup)
        code_format = "{{code{title}}}\n{content}\n{{code}}"
        tags = {}
        for idx, code in enumerate(soup.findAll("code")):
            tag = self.tag_format.format(idx=str(idx))
            title = code.attrs.get("class", "")
            if type(title) == list:
                title = title[0] if len(title) > 0 else ""
            title = title.replace("language-", "")
            if title.lower() in ["burp", "http", "shell", "graphql"]:
                title = ""
            tags[tag] = code_format.format(
                title=":{title}".format(title=title) if title else "",
                content="".join([str(i) for i in code.contents]),
            )
            n_html = n_html.replace(str(code), "{%(tag)s}" % ({"tag": tag}))
        return unescape_text(n_html), tags

    def _img_to_jira_tag(self, html):
        """
        Replace "img" balisis in html to "a" balisis.
        """
        soup = BeautifulSoup(html, features="lxml")
        n_html = str(soup)
        for img in soup.findAll("img"):
            alt = img.attrs.get("alt", "")
            src = img.attrs.get("src", "")
            a = f"!{alt}|{src}!"
            n_html = n_html.replace(str(img), a)
        return unescape_text(n_html)


class YWHJiraConfig(BugTrackerConfig):

    """
    BugTrackerConfig Class associated to jira.

    :attr str bugtracker_type: bugtracker identifier
    :attr class client: Client class
    :attr list mandatory_keys: keys needed in configuration file part
    :attr list secret_keys: keys need interaction in interactive mode
    :attr dict optional_keys: keys with default values
    :attr dict _description: descriptor for each specified key
    """

    bugtracker_type = "jira"
    client = YWHJira

    mandatory_keys = ["url", "login", "project"]
    secret_keys = ["password"]
    optional_keys = dict(issuetype="Task")
    _description = dict(project="Jira slug")

    ############################################################
    #################### Instance methods ######################
    ############################################################
    def _set_bugtracker(self):
        self._get_bugtracker(
            self._url,
            self._login,
            self._password,
            self._project,
            issuetype=self._issuetype,
        )
